@@ -2,6 +2,7 @@
 {-# LANGUAGE ScopedTypeVariables #-}
 {-# LANGUAGE TypeFamilies #-}
 {-# LANGUAGE RankNTypes #-}
+{-# LANGUAGE TemplateHaskellQuotes #-}
 
 -- |
 -- Module      : Data.Primitive.ByteArray
@@ -24,14 +25,13 @@ module Data.Primitive.ByteArray (
   -- * Allocation
   newByteArray, newPinnedByteArray, newAlignedPinnedByteArray,
   resizeMutableByteArray,
-#if __GLASGOW_HASKELL__ >= 710
   shrinkMutableByteArray,
-#endif
 
   -- * Element access
   readByteArray, writeByteArray, indexByteArray,
 
   -- * Constructing
+  emptyByteArray,
   byteArrayFromList, byteArrayFromListN,
 
   -- * Folding
@@ -41,15 +41,14 @@ module Data.Primitive.ByteArray (
   compareByteArrays,
 
   -- * Freezing and thawing
-  freezeByteArray,
+  freezeByteArray, thawByteArray, runByteArray,
   unsafeFreezeByteArray, unsafeThawByteArray,
 
   -- * Block operations
   copyByteArray, copyMutableByteArray,
-#if __GLASGOW_HASKELL__ >= 708
   copyByteArrayToPtr, copyMutableByteArrayToPtr,
   copyByteArrayToAddr, copyMutableByteArrayToAddr,
-#endif
+  copyPtrToMutableByteArray,
   moveByteArray,
   setByteArray, fillByteArray,
   cloneByteArray, cloneMutableByteArray,
@@ -67,52 +66,87 @@ module Data.Primitive.ByteArray (
 import Control.Monad.Primitive
 import Control.Monad.ST
 import Control.DeepSeq
-import Data.Data (mkNoRepType)
 import Data.Primitive.Types
 
 import qualified GHC.ST as GHCST
 
 import Foreign.C.Types
 import Data.Word ( Word8 )
-import GHC.Base ( Int(..) )
-#if __GLASGOW_HASKELL__ >= 708
-import qualified GHC.Exts as Exts ( IsList(..) )
-#endif
-import GHC.Exts
-#if __GLASGOW_HASKELL__ >= 706
-    hiding (setByteArray#)
-#endif
+import Data.Bits ( (.&.), unsafeShiftR )
+import GHC.Show ( intToDigit )
+import qualified GHC.Exts as Exts
+import GHC.Exts hiding (setByteArray#)
 
 import Data.Typeable ( Typeable )
-import Data.Data ( Data(..) )
-import Data.Primitive.Internal.Compat ( isTrue# )
-import Numeric
+import Data.Data ( Data(..), mkNoRepType )
+import qualified Language.Haskell.TH.Syntax as TH
+import qualified Language.Haskell.TH.Lib as TH
 
-#if MIN_VERSION_base(4,9,0)
 import qualified Data.Semigroup as SG
 import qualified Data.Foldable as F
-#endif
 
-#if !(MIN_VERSION_base(4,8,0))
-import Data.Monoid (Monoid(..))
-#endif
+import System.IO.Unsafe (unsafePerformIO, unsafeDupablePerformIO)
 
-#if __GLASGOW_HASKELL__ >= 802
-import GHC.Exts as Exts (isByteArrayPinned#,isMutableByteArrayPinned#)
-#endif
-
-#if __GLASGOW_HASKELL__ >= 804
-import GHC.Exts (compareByteArrays#)
-#else
-import System.IO.Unsafe (unsafeDupablePerformIO)
-#endif
-
--- | Byte arrays
+-- | Byte arrays.
 data ByteArray = ByteArray ByteArray# deriving ( Typeable )
 
--- | Mutable byte arrays associated with a primitive state token
+-- | Mutable byte arrays associated with a primitive state token.
 data MutableByteArray s = MutableByteArray (MutableByteArray# s)
-                                        deriving( Typeable )
+  deriving ( Typeable )
+
+-- | Respects array pinnedness for GHC >= 8.2
+instance TH.Lift ByteArray where
+#if MIN_VERSION_template_haskell(2,17,0)
+  liftTyped ba = TH.unsafeCodeCoerce (TH.lift ba)
+#elif MIN_VERSION_template_haskell(2,16,0)
+  liftTyped ba = TH.unsafeTExpCoerce (TH.lift ba)
+#endif
+
+  lift ba =
+    TH.appE
+      (if small
+         then [| fromLitAddrSmall# pinned len |]
+         else [| fromLitAddrLarge# pinned len |])
+      (TH.litE (TH.stringPrimL (toList ba)))
+    where
+      -- Pin it if the original was pinned; otherwise don't. This seems more
+      -- logical to me than the alternatives. Anyone who wants a different
+      -- pinnedness can just copy the compile-time byte array to one that
+      -- matches what they want at run-time.
+#if __GLASGOW_HASKELL__ >= 802
+      pinned = isByteArrayPinned ba
+#else
+      pinned = True
+#endif
+      len = sizeofByteArray ba
+      small = len <= 2048
+
+-- I don't think inlining these can be very helpful, so let's not
+-- do it.
+{-# NOINLINE fromLitAddrSmall# #-}
+fromLitAddrSmall# :: Bool -> Int -> Addr# -> ByteArray
+fromLitAddrSmall# pinned len ptr = inline (fromLitAddr# True pinned len ptr)
+
+{-# NOINLINE fromLitAddrLarge# #-}
+fromLitAddrLarge# :: Bool -> Int -> Addr# -> ByteArray
+fromLitAddrLarge# pinned len ptr = inline (fromLitAddr# False pinned len ptr)
+
+fromLitAddr# :: Bool -> Bool -> Int -> Addr# -> ByteArray
+fromLitAddr# small pinned !len !ptr = upIO $ do
+  mba <- if pinned
+         then newPinnedByteArray len
+         else newByteArray len
+  copyPtrToMutableByteArray mba 0 (Ptr ptr :: Ptr Word8) len
+  unsafeFreezeByteArray mba
+  where
+    -- We don't care too much about duplication if the byte arrays are
+    -- small. If they're large, we do. Since we don't allocate while
+    -- we copy (we do it with a primop!), I don't believe the thunk
+    -- deduplication mechanism can help us if two threads just happen
+    -- to try to build the ByteArray at the same time.
+    upIO
+      | small = unsafeDupablePerformIO
+      | otherwise = unsafePerformIO
 
 instance NFData ByteArray where
   rnf (ByteArray _) = ()
@@ -190,16 +224,9 @@ resizeMutableByteArray
   :: PrimMonad m => MutableByteArray (PrimState m) -> Int
                  -> m (MutableByteArray (PrimState m))
 {-# INLINE resizeMutableByteArray #-}
-#if __GLASGOW_HASKELL__ >= 710
 resizeMutableByteArray (MutableByteArray arr#) (I# n#)
   = primitive (\s# -> case resizeMutableByteArray# arr# n# s# of
                         (# s'#, arr'# #) -> (# s'#, MutableByteArray arr'# #))
-#else
-resizeMutableByteArray arr n
-  = do arr' <- newByteArray n
-       copyMutableByteArray arr' 0 arr 0 (min (sizeofMutableByteArray arr) n)
-       return arr'
-#endif
 
 -- | Get the size of a byte array in bytes. Unlike 'sizeofMutableByteArray',
 -- this function ensures sequencing in the presence of resizing.
@@ -220,6 +247,9 @@ getSizeofMutableByteArray arr
 --
 -- This operation makes a copy of the specified section, so it is safe to
 -- continue using the mutable array afterward.
+--
+-- /Note:/ The provided array should contain the full subrange
+-- specified by the two Ints, but this is not checked.
 freezeByteArray
   :: PrimMonad m
   => MutableByteArray (PrimState m) -- ^ source
@@ -232,6 +262,28 @@ freezeByteArray !src !off !len = do
   copyMutableByteArray dst 0 src off len
   unsafeFreezeByteArray dst
 
+-- | Create a mutable byte array from a slice of an immutable byte array.
+-- The offset and length are given in bytes.
+--
+-- This operation makes a copy of the specified slice, so it is safe to
+-- use the immutable array afterward.
+--
+-- /Note:/ The provided array should contain the full subrange
+-- specified by the two Ints, but this is not checked.
+--
+-- @since 0.7.2.0
+thawByteArray
+  :: PrimMonad m
+  => ByteArray -- ^ source
+  -> Int       -- ^ offset in bytes
+  -> Int       -- ^ length in bytes
+  -> m (MutableByteArray (PrimState m))
+{-# INLINE thawByteArray #-}
+thawByteArray !src !off !len = do
+  dst <- newByteArray len
+  copyByteArray dst 0 src off len
+  return dst
+
 -- | Convert a mutable byte array to an immutable one without copying. The
 -- array should not be modified after the conversion.
 unsafeFreezeByteArray
@@ -263,12 +315,8 @@ sizeofMutableByteArray :: MutableByteArray s -> Int
 {-# INLINE sizeofMutableByteArray #-}
 sizeofMutableByteArray (MutableByteArray arr#) = I# (sizeofMutableByteArray# arr#)
 
--- Although it is possible to shim resizeMutableByteArray for old GHCs, this
--- is not the case with shrinkMutableByteArray.
-#if __GLASGOW_HASKELL__ >= 710
 -- | Shrink a mutable byte array. The new size is given in bytes.
 -- It must be smaller than the old size. The array will be resized in place.
--- This function is only available when compiling with GHC 7.10 or newer.
 --
 -- @since 0.7.1.0
 shrinkMutableByteArray :: PrimMonad m
@@ -278,23 +326,26 @@ shrinkMutableByteArray :: PrimMonad m
 {-# INLINE shrinkMutableByteArray #-}
 shrinkMutableByteArray (MutableByteArray arr#) (I# n#)
   = primitive_ (shrinkMutableByteArray# arr# n#)
-#endif
 
 #if __GLASGOW_HASKELL__ >= 802
 -- | Check whether or not the byte array is pinned. Pinned byte arrays cannot
---   be moved by the garbage collector. It is safe to use 'byteArrayContents'
---   on such byte arrays. This function is only available when compiling with
---   GHC 8.2 or newer.
+-- be moved by the garbage collector. It is safe to use 'byteArrayContents' on
+-- such byte arrays.
 --
---   @since 0.6.4.0
+-- Caution: This function is only available when compiling with GHC 8.2 or
+-- newer.
+--
+-- @since 0.6.4.0
 isByteArrayPinned :: ByteArray -> Bool
 {-# INLINE isByteArrayPinned #-}
 isByteArrayPinned (ByteArray arr#) = isTrue# (Exts.isByteArrayPinned# arr#)
 
--- | Check whether or not the mutable byte array is pinned. This function is
---   only available when compiling with GHC 8.2 or newer.
+-- | Check whether or not the mutable byte array is pinned.
 --
---   @since 0.6.4.0
+-- Caution: This function is only available when compiling with GHC 8.2 or
+-- newer.
+--
+-- @since 0.6.4.0
 isMutableByteArrayPinned :: MutableByteArray s -> Bool
 {-# INLINE isMutableByteArrayPinned #-}
 isMutableByteArrayPinned (MutableByteArray marr#) = isTrue# (Exts.isMutableByteArrayPinned# marr#)
@@ -334,7 +385,7 @@ foldrByteArray :: forall a b. (Prim a) => (a -> b -> b) -> b -> ByteArray -> b
 foldrByteArray f z arr = go 0
   where
     go i
-      | i < maxI  = f (indexByteArray arr i) (go (i+1))
+      | i < maxI  = f (indexByteArray arr i) (go (i + 1))
       | otherwise = z
     maxI = sizeofByteArray arr `quot` sizeOf (undefined :: a)
 
@@ -345,7 +396,7 @@ byteArrayFromList :: Prim a => [a] -> ByteArray
 byteArrayFromList xs = byteArrayFromListN (length xs) xs
 
 -- | Create a 'ByteArray' from a list of a known length. If the length
---   of the list does not match the given length, this throws an exception.
+-- of the list does not match the given length, this throws an exception.
 byteArrayFromListN :: Prim a => Int -> [a] -> ByteArray
 byteArrayFromListN n ys = runST $ do
     marr <- newByteArray (n * sizeOf (head ys))
@@ -367,13 +418,13 @@ unI# (I# n#) = n#
 --
 -- /Note:/ this function does not do bounds or overlap checking.
 copyByteArray
-  :: PrimMonad m => MutableByteArray (PrimState m)
-                                        -- ^ destination array
-                 -> Int                 -- ^ offset into destination array
-                 -> ByteArray           -- ^ source array
-                 -> Int                 -- ^ offset into source array
-                 -> Int                 -- ^ number of bytes to copy
-                 -> m ()
+  :: PrimMonad m
+  => MutableByteArray (PrimState m) -- ^ destination array
+  -> Int                            -- ^ offset into destination array
+  -> ByteArray                      -- ^ source array
+  -> Int                            -- ^ offset into source array
+  -> Int                            -- ^ number of bytes to copy
+  -> m ()
 {-# INLINE copyByteArray #-}
 copyByteArray (MutableByteArray dst#) doff (ByteArray src#) soff sz
   = primitive_ (copyByteArray# src# (unI# soff) dst# (unI# doff) (unI# sz))
@@ -383,27 +434,24 @@ copyByteArray (MutableByteArray dst#) doff (ByteArray src#) soff sz
 --
 -- /Note:/ this function does not do bounds or overlap checking.
 copyMutableByteArray
-  :: PrimMonad m => MutableByteArray (PrimState m)
-                                        -- ^ destination array
-                 -> Int                 -- ^ offset into destination array
-                 -> MutableByteArray (PrimState m)
-                                        -- ^ source array
-                 -> Int                 -- ^ offset into source array
-                 -> Int                 -- ^ number of bytes to copy
-                 -> m ()
+  :: PrimMonad m
+  => MutableByteArray (PrimState m) -- ^ destination array
+  -> Int                            -- ^ offset into destination array
+  -> MutableByteArray (PrimState m) -- ^ source array
+  -> Int                            -- ^ offset into source array
+  -> Int                            -- ^ number of bytes to copy
+  -> m ()
 {-# INLINE copyMutableByteArray #-}
 copyMutableByteArray (MutableByteArray dst#) doff
                      (MutableByteArray src#) soff sz
   = primitive_ (copyMutableByteArray# src# (unI# soff) dst# (unI# doff) (unI# sz))
 
-#if __GLASGOW_HASKELL__ >= 708
--- | Copy a slice of a byte array to an unmanaged Pointer Address. These must not
---   overlap. The offset and length given in elements, not in bytes. This function
---   is only available when compiling with GHC 7.8 or newer.
+-- | Copy a slice of a byte array to an unmanaged pointer address. These must not
+-- overlap. The offset and length are given in elements, not in bytes.
 --
---   /Note:/ this function does not do bounds or overlap checking.
+-- /Note:/ this function does not do bounds or overlap checking.
 --
---   @since 0.7.1.0
+-- @since 0.7.1.0
 copyByteArrayToPtr
   :: forall m a. (PrimMonad m, Prim a)
   => Ptr a -- ^ destination
@@ -417,14 +465,30 @@ copyByteArrayToPtr (Ptr dst#) (ByteArray src#) soff sz
   where
   siz# = sizeOf# (undefined :: a)
 
--- | Copy a slice of a mutable byte array to an unmanaged Pointer address.
---   These must not overlap. The offset and length given in elements, not
---   in bytes. This function is only available when compiling with GHC 7.8
---   or newer.
+-- | Copy from an unmanaged pointer address to a byte array. These must not
+-- overlap. The offset and length are given in elements, not in bytes.
 --
---   /Note:/ this function does not do bounds or overlap checking.
+-- /Note:/ this function does not do bounds or overlap checking.
+copyPtrToMutableByteArray :: forall m a. (PrimMonad m, Prim a)
+  => MutableByteArray (PrimState m) -- ^ destination array
+  -> Int   -- ^ destination offset given in elements of type @a@
+  -> Ptr a -- ^ source pointer
+  -> Int   -- ^ number of elements
+  -> m ()
+{-# INLINE copyPtrToMutableByteArray #-}
+copyPtrToMutableByteArray (MutableByteArray ba#) (I# doff#) (Ptr addr#) (I# n#) =
+  primitive_ (copyAddrToByteArray# addr# ba# (doff# *# siz#) (n# *# siz#))
+  where
+  siz# = sizeOf# (undefined :: a)
+
+
+-- | Copy a slice of a mutable byte array to an unmanaged pointer address.
+-- These must not overlap. The offset and length are given in elements, not
+-- in bytes.
 --
---   @since 0.7.1.0
+-- /Note:/ this function does not do bounds or overlap checking.
+--
+-- @since 0.7.1.0
 copyMutableByteArrayToPtr
   :: forall m a. (PrimMonad m, Prim a)
   => Ptr a -- ^ destination
@@ -443,12 +507,11 @@ copyMutableByteArrayToPtr (Ptr dst#) (MutableByteArray src#) soff sz
 -----
 
 -- | Copy a slice of a byte array to an unmanaged address. These must not
---   overlap. This function is only available when compiling with GHC 7.8
---   or newer.
+-- overlap.
 --
---   Note: This function is just 'copyByteArrayToPtr' where @a@ is 'Word8'.
+-- Note: This function is just 'copyByteArrayToPtr' where @a@ is 'Word8'.
 --
---   @since 0.6.4.0
+-- @since 0.6.4.0
 copyByteArrayToAddr
   :: PrimMonad m
   => Ptr Word8 -- ^ destination
@@ -461,12 +524,11 @@ copyByteArrayToAddr (Ptr dst#) (ByteArray src#) soff sz
   = primitive_ (copyByteArrayToAddr# src# (unI# soff) dst# (unI# sz))
 
 -- | Copy a slice of a mutable byte array to an unmanaged address. These must
---   not overlap. This function is only available when compiling with GHC 7.8
---   or newer.
+-- not overlap.
 --
---   Note: This function is just 'copyMutableByteArrayToPtr' where @a@ is 'Word8'.
+-- Note: This function is just 'copyMutableByteArrayToPtr' where @a@ is 'Word8'.
 --
---   @since 0.6.4.0
+-- @since 0.6.4.0
 copyMutableByteArrayToAddr
   :: PrimMonad m
   => Ptr Word8 -- ^ destination
@@ -477,19 +539,17 @@ copyMutableByteArrayToAddr
 {-# INLINE copyMutableByteArrayToAddr #-}
 copyMutableByteArrayToAddr (Ptr dst#) (MutableByteArray src#) soff sz
   = primitive_ (copyMutableByteArrayToAddr# src# (unI# soff) dst# (unI# sz))
-#endif
 
 -- | Copy a slice of a mutable byte array into another, potentially
 -- overlapping array.
 moveByteArray
-  :: PrimMonad m => MutableByteArray (PrimState m)
-                                        -- ^ destination array
-                 -> Int                 -- ^ offset into destination array
-                 -> MutableByteArray (PrimState m)
-                                        -- ^ source array
-                 -> Int                 -- ^ offset into source array
-                 -> Int                 -- ^ number of bytes to copy
-                 -> m ()
+  :: PrimMonad m
+  => MutableByteArray (PrimState m) -- ^ destination array
+  -> Int                            -- ^ offset into destination array
+  -> MutableByteArray (PrimState m) -- ^ source array
+  -> Int                            -- ^ offset into source array
+  -> Int                            -- ^ number of bytes to copy
+  -> m ()
 {-# INLINE moveByteArray #-}
 moveByteArray (MutableByteArray dst#) doff
               (MutableByteArray src#) soff sz
@@ -502,11 +562,12 @@ moveByteArray (MutableByteArray dst#) doff
 --
 -- /Note:/ this function does not do bounds checking.
 setByteArray
-  :: (Prim a, PrimMonad m) => MutableByteArray (PrimState m) -- ^ array to fill
-                           -> Int                 -- ^ offset into array
-                           -> Int                 -- ^ number of values to fill
-                           -> a                   -- ^ value to fill with
-                           -> m ()
+  :: (Prim a, PrimMonad m)
+  => MutableByteArray (PrimState m) -- ^ array to fill
+  -> Int                            -- ^ offset into array
+  -> Int                            -- ^ number of values to fill
+  -> a                              -- ^ value to fill with
+  -> m ()
 {-# INLINE setByteArray #-}
 setByteArray (MutableByteArray dst#) (I# doff#) (I# sz#) x
   = primitive_ (setByteArray# dst# doff# sz# x)
@@ -515,12 +576,12 @@ setByteArray (MutableByteArray dst#) (I# doff#) (I# sz#) x
 --
 -- /Note:/ this function does not do bounds checking.
 fillByteArray
-  :: PrimMonad m => MutableByteArray (PrimState m)
-                                        -- ^ array to fill
-                 -> Int                 -- ^ offset into array
-                 -> Int                 -- ^ number of bytes to fill
-                 -> Word8               -- ^ byte to fill with
-                 -> m ()
+  :: PrimMonad m
+  => MutableByteArray (PrimState m) -- ^ array to fill
+  -> Int                            -- ^ offset into array
+  -> Int                            -- ^ number of bytes to fill
+  -> Word8                          -- ^ byte to fill with
+  -> m ()
 {-# INLINE fillByteArray #-}
 fillByteArray = setByteArray
 
@@ -543,12 +604,24 @@ instance Typeable s => Data (MutableByteArray s) where
   dataTypeOf _ = mkNoRepType "Data.Primitive.ByteArray.MutableByteArray"
 
 -- | @since 0.6.3.0
+--
+-- Behavior changed in 0.7.2.0. Before 0.7.2.0, this instance rendered
+-- 8-bit words less than 16 as a single hexadecimal digit (e.g. 13 was @0xD@).
+-- Starting with 0.7.2.0, all 8-bit words are represented as two digits
+-- (e.g. 13 is @0x0D@).
 instance Show ByteArray where
   showsPrec _ ba =
       showString "[" . go 0
     where
+      showW8 :: Word8 -> String -> String
+      showW8 !w s =
+          '0'
+        : 'x'
+        : intToDigit (fromIntegral (unsafeShiftR w 4))
+        : intToDigit (fromIntegral (w .&. 0x0F))
+        : s
       go i
-        | i < sizeofByteArray ba = comma . showString "0x" . showHex (indexByteArray ba i :: Word8) . go (i+1)
+        | i < sizeofByteArray ba = comma . showW8 (indexByteArray ba i :: Word8) . go (i+1)
         | otherwise              = showChar ']'
         where
           comma | i == 0    = id
@@ -575,12 +648,12 @@ foreign import ccall unsafe "primitive-memops.h hsprimitive_memcmp"
 
 -- | Lexicographic comparison of equal-length slices into two byte arrays.
 -- This wraps the @compareByteArrays#@ primop, which wraps @memcmp@.
-compareByteArrays ::
-     ByteArray -- ^ Array A
-  -> Int -- ^ Offset A, given in bytes
-  -> ByteArray -- ^ Array B
-  -> Int -- ^ Offset B, given in bytes
-  -> Int -- ^ Length of slice, given in bytes
+compareByteArrays
+  :: ByteArray -- ^ array A
+  -> Int       -- ^ offset A, given in bytes
+  -> ByteArray -- ^ array B
+  -> Int       -- ^ offset B, given in bytes
+  -> Int       -- ^ length of the slice, given in bytes
   -> Ordering
 {-# INLINE compareByteArrays #-}
 #if __GLASGOW_HASKELL__ >= 804
@@ -602,12 +675,7 @@ foreign import ccall unsafe "primitive-memops.h hsprimitive_memcmp_offset"
 sameByteArray :: ByteArray# -> ByteArray# -> Bool
 sameByteArray ba1 ba2 =
     case reallyUnsafePtrEquality# (unsafeCoerce# ba1 :: ()) (unsafeCoerce# ba2 :: ()) of
-#if __GLASGOW_HASKELL__ >= 708
       r -> isTrue# r
-#else
-      1# -> True
-      0# -> False
-#endif
 
 -- | @since 0.6.3.0
 instance Eq ByteArray where
@@ -663,7 +731,9 @@ calcLength :: [ByteArray] -> Int -> Int
 calcLength [] !n = n
 calcLength (x : xs) !n = calcLength xs (sizeofByteArray x + n)
 
+-- | The empty 'ByteArray'.
 emptyByteArray :: ByteArray
+{-# NOINLINE emptyByteArray #-}
 emptyByteArray = runST (newByteArray 0 >>= unsafeFreezeByteArray)
 
 replicateByteArray :: Int -> ByteArray -> ByteArray
@@ -677,16 +747,13 @@ replicateByteArray n arr = runST $ do
   go 0
   unsafeFreezeByteArray marr
 
-#if MIN_VERSION_base(4,9,0)
 instance SG.Semigroup ByteArray where
   (<>) = appendByteArray
   sconcat = mconcat . F.toList
-  stimes i arr
-    | itgr < 1 = emptyByteArray
-    | itgr <= (fromIntegral (maxBound :: Int)) = replicateByteArray (fromIntegral itgr) arr
-    | otherwise = error "Data.Primitive.ByteArray#stimes: cannot allocate the requested amount of memory"
-    where itgr = toInteger i :: Integer
-#endif
+  stimes n arr = case compare n 0 of
+    LT -> die "stimes" "negative multiplier"
+    EQ -> emptyByteArray
+    GT -> replicateByteArray (fromIntegral n) arr
 
 instance Monoid ByteArray where
   mempty = emptyByteArray
@@ -695,7 +762,6 @@ instance Monoid ByteArray where
 #endif
   mconcat = concatByteArray
 
-#if __GLASGOW_HASKELL__ >= 708
 -- | @since 0.6.3.0
 instance Exts.IsList ByteArray where
   type Item ByteArray = Word8
@@ -703,7 +769,6 @@ instance Exts.IsList ByteArray where
   toList = foldrByteArray (:) []
   fromList xs = byteArrayFromListN (length xs) xs
   fromListN = byteArrayFromListN
-#endif
 
 die :: String -> String -> a
 die fun problem = error $ "Data.Primitive.ByteArray." ++ fun ++ ": " ++ problem
@@ -711,8 +776,8 @@ die fun problem = error $ "Data.Primitive.ByteArray." ++ fun ++ ": " ++ problem
 -- | Return a newly allocated array with the specified subrange of the
 -- provided array. The provided array should contain the full subrange
 -- specified by the two Ints, but this is not checked.
-cloneByteArray ::
-     ByteArray -- ^ source array
+cloneByteArray
+  :: ByteArray -- ^ source array
   -> Int       -- ^ offset into destination array
   -> Int       -- ^ number of bytes to copy
   -> ByteArray
@@ -736,10 +801,13 @@ cloneMutableByteArray src off n = do
   copyMutableByteArray dst 0 src off n
   return dst
 
-#if MIN_VERSION_base(4,10,0) /* In new GHCs, runRW# is available. */
+-- | Execute the monadic action and freeze the resulting array.
+--
+-- > runByteArray m = runST $ m >>= unsafeFreezeByteArray
 runByteArray
   :: (forall s. ST s (MutableByteArray s))
   -> ByteArray
+#if MIN_VERSION_base(4,10,0) /* In new GHCs, runRW# is available. */
 runByteArray m = ByteArray (runByteArray# m)
 
 runByteArray#
@@ -752,8 +820,5 @@ runByteArray# m = case runRW# $ \s ->
 unST :: ST s a -> State# s -> (# State# s, a #)
 unST (GHCST.ST f) = f
 #else /* In older GHCs, runRW# is not available. */
-runByteArray
-  :: (forall s. ST s (MutableByteArray s))
-  -> ByteArray
 runByteArray m = runST $ m >>= unsafeFreezeByteArray
 #endif
